"""Sentinels used internally by the Viv compiler."""

from typing import Final

from viv_compiler import internal_types

# Sentinel used to mark the child-join operator
CHILD_JOIN_OPERATOR: Final[internal_types.Sentinel] = object()

# Sentinel used to mark an entity binding type
ENTITY_BINDING_TYPE: Final[internal_types.Sentinel] = object()

# Sentinel used to mark a symbol binding type
SYMBOL_BINDING_TYPE: Final[internal_types.Sentinel] = object()

# Sentinel used to represent the negation of an expression
NEGATION: Final[internal_types.Sentinel] = object()

# Sentinel for an action temporal anchor in time-frame statements
ACTION_TEMPORAL_ANCHOR: Final[internal_types.Sentinel] = object()

# Sentinel for a non-action temporal anchor in time-frame statements
NON_ACTION_TEMPORAL_ANCHOR: Final[internal_types.Sentinel] = object()

# Sentinel for the entity sigil
ENTITY_SIGIL: Final[internal_types.Sentinel] = object()

# Sentinel for the symbol sigil
SYMBOL_SIGIL: Final[internal_types.Sentinel] = object()

# Sentinel for the local-variable sigil
LOCAL_VARIABLE_SIGIL: Final[internal_types.Sentinel] = object()

# Sentinel for the scratch-variable sigil
SCRATCH_VARIABLE_SIGIL: Final[internal_types.Sentinel] = object()

# Sentinel for the plan-phase sigil
PLAN_PHASE_SIGIL: Final[internal_types.Sentinel] = object()

# Sentinel for the group-role decorator
GROUP_ROLE_DECORATOR: Final[internal_types.Sentinel] = object()

# Sentinel for the eval-fail-safe marker
EVAL_FAIL_SAFE_MARKER: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the tags of a child and parent action
CHILD_JOIN_DIRECTIVE_TAGS: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the roles of a child and parent action
CHILD_JOIN_DIRECTIVE_ROLES: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the conditions of a child and parent action
CHILD_JOIN_DIRECTIVE_CONDITIONS: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the scratch field of a child and parent action
CHILD_JOIN_DIRECTIVE_SCRATCH: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the effects of a child and parent action
CHILD_JOIN_DIRECTIVE_EFFECTS: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the reactions of a child and parent action
CHILD_JOIN_DIRECTIVE_REACTIONS: Final[internal_types.Sentinel] = object()

# Sentinel for a directive to join the embargoes of a child and parent action
CHILD_JOIN_DIRECTIVE_EMBARGOES: Final[internal_types.Sentinel] = object()

# Sentinel for the permanent embargo marker
PERMANENT_EMBARGO_MARKER: Final[internal_types.Sentinel] = object()

# Sentinel for the location-specific embargo marker
LOCATION_SPECIFIC_EMBARGO_MARKER: Final[internal_types.Sentinel] = object()

# Sentinel for the location-agnostic embargo marker
LOCATION_AGNOSTIC_EMBARGO_MARKER: Final[internal_types.Sentinel] = object()

# Sentinel marking that a reaction targets an action
REACTION_TARGET_TYPE_ACTION: Final[internal_types.Sentinel] = object()

# Sentinel marking that a reaction targets an action selector
REACTION_TARGET_TYPE_ACTION_SELECTOR: Final[internal_types.Sentinel] = object()

# Sentinel marking that a reaction targets a plan
REACTION_TARGET_TYPE_PLAN: Final[internal_types.Sentinel] = object()

# Sentinel marking that a reaction targets a plan selector
REACTION_TARGET_TYPE_PLAN_SELECTOR: Final[internal_types.Sentinel] = object()

# Sentinel marking an action search with no target query
ACTION_SEARCH_NO_QUERY: Final[internal_types.Sentinel] = object()

# Sentinel marking that a search domain (for an action search or sifting) is the full chronicle
SEARCH_DOMAIN_CHRONICLE: Final[internal_types.Sentinel] = object()

# Sentinel marking that a search domain (for an action search or sifting)
# is to be inherited from the enclosing context.
SEARCH_DOMAIN_INHERIT: Final[internal_types.Sentinel] = object()

# Sentinel for the 'all' reaction-window operator used in plans
REACTION_WINDOW_OPERATOR_ALL: Final[internal_types.Sentinel] = object()

# Sentinel for the 'any' reaction-window operator used in plans
REACTION_WINDOW_OPERATOR_ANY: Final[internal_types.Sentinel] = object()

# Sentinel for the 'untracked' reaction-window operator used in plans
REACTION_WINDOW_OPERATOR_UNTRACKED: Final[internal_types.Sentinel] = object()

# Sentinel for a randomized selector policy
SELECTOR_POLICY_RANDOM: Final[internal_types.Sentinel] = object()

# Sentinel for a weighted selector policy
SELECTOR_POLICY_WEIGHTED: Final[internal_types.Sentinel] = object()

# Sentinel for an ordered selector policy
SELECTOR_POLICY_ORDERED: Final[internal_types.Sentinel] = object()



"""Mixin class containing visitor methods associated with plan definitions."""

__all__ = ["VisitorMixinPlans"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, sentinels, utils, validation


class VisitorMixinPlans(PTNodeVisitor):
    """A visitor mixin for Viv plans."""

    @staticmethod
    def visit_plan(node: NonTerminal, children: list[Any]) -> internal_types.IntermediatePlanDefinition:
        """Visit a <plan> node."""
        # Check for required fields
        accumulated_fields = {}
        plan_name, plan_body = children
        accumulated_fields['name'] = plan_name
        if "roles" not in plan_body:
            error_message = f"Plan '{accumulated_fields['name']}' is missing 'roles' field (required)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if "phases" not in plan_body:
            error_message = f"Plan '{accumulated_fields['name']}' is missing 'phases' field (required)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        # Collect all role definitions
        all_role_definitions = plan_body["roles"]
        validation.prevalidate_role_names(
            construct_type=external_types.ConstructDiscriminator.PLAN,
            construct_name=plan_name,
            role_definitions=all_role_definitions,
            source=utils.derive_source_annotations(node=node)
        )
        accumulated_fields["roles"] = {role["name"]: role for role in all_role_definitions}
        # Collect all phase definitions, and set the `next` property for each
        all_phase_definitions = plan_body["phases"]
        for i in range(len(all_phase_definitions) - 1):
            all_phase_definitions[i]['next'] = all_phase_definitions[i + 1]['name']
        all_phase_definitions[len(all_phase_definitions) - 1]['next'] = None
        accumulated_fields["phases"] = {phase["name"]: phase for phase in all_phase_definitions}
        accumulated_fields["initialPhase"] = all_phase_definitions[0]['name']
        # Populate the fields
        for field_name, field_value in plan_body.items():
            if field_name == "roles" or field_name == "phases":
                continue  # Already handled just above
            else:
                accumulated_fields[field_name] = field_value
        # Do some quick prevalidation that depends on transient data
        for phase_definition in all_phase_definitions:
            for instruction in phase_definition["tape"]:
                if instruction["type"] == external_types.PlanInstructionDiscriminator.REACTION_QUEUE:
                    instruction: external_types.PlanInstructionReactionQueue
                    reaction = instruction["reaction"]["value"]
                    if not reaction['time']:
                        continue
                    temporal_statements = reaction['time']
                    for temporal_statement in temporal_statements:
                        if temporal_statement.get("useActionTimestamp"):
                            error_message = (
                                f"Plan '{plan_name}' has phase '{phase_definition['name']}' with temporal "
                                f"constraint with bad anchor 'from action' ('from now' is required in plans)"
                            )
                            raise errors.VivCompileError(
                                error_message,
                                source=utils.derive_source_annotations(node=node)
                            )
        # Package up the intermediate definition
        intermediate_plan_definition = internal_types.IntermediatePlanDefinition(
            type=external_types.ConstructDiscriminator.PLAN,
            name=plan_name,
            roles=accumulated_fields["roles"],
            _conditions_raw=accumulated_fields.get("_conditions_raw", []),
            phases=accumulated_fields["phases"],
            initialPhase=accumulated_fields["initialPhase"]
        )
        return intermediate_plan_definition

    @staticmethod
    def visit_plan_header(_, children: list[Any]) -> external_types.PlanName:
        """Visit a <plan_header> node."""
        return children[0]

    @staticmethod
    def visit_plan_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <plan_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_plan_roles(_, children: list[Any]) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <plan_roles> node."""
        return {"roles": children}

    @staticmethod
    def visit_plan_conditions(_, children: list[Any]) -> dict[str, list[external_types.Expression]]:
        """Visit a <plan_conditions> node."""
        return {"_conditions_raw": children[0]}

    @staticmethod
    def visit_plan_phases(_, children: list[Any]) -> dict[str, list[external_types.PlanPhase]]:
        """Visit a <plan_phases> node."""
        return {"phases": children}

    @staticmethod
    def visit_plan_phase(_, children: list[Any]) -> external_types.PlanPhase:
        """Visit a <plan_phase> node."""
        phase_name, phase_tape = children
        phase_definition = external_types.PlanPhase(
            name=phase_name,
            next=None,  # Set in `visit_plan()`, once all the phases have been compiled
            tape=phase_tape
        )
        return phase_definition

    @staticmethod
    def visit_plan_phase_name(_, children: list[Any]) -> external_types.PlanPhaseName:
        """Visit a <plan_phase_name> node."""
        return children[-1]

    @staticmethod
    def visit_plan_instructions(
        node: NonTerminal,
        children: list[list[internal_types.IntermediatePlanInstruction]]
    ) -> list[external_types.PlanInstruction]:
        """Visit a <plan_instructions> node."""
        # Linearize the raw tape, whose instructions may include unresolved offset addresses
        raw_tape: list[internal_types.IntermediatePlanInstruction] = []
        for intermediate_instructions_group in children:
            raw_tape += intermediate_instructions_group
        # Resolve all address offsets (into actual addresses)
        final_tape: list[external_types.PlanInstruction] = []
        for instruction_address, instruction in enumerate(raw_tape):
            instruction: internal_types.IntermediatePlanInstruction
            match instruction['type']:
                case(
                    external_types.PlanInstructionDiscriminator.ADVANCE
                    | external_types.PlanInstructionDiscriminator.FAIL
                    | external_types.PlanInstructionDiscriminator.LOOP_INIT
                    | external_types.PlanInstructionDiscriminator.REACTION_QUEUE
                    | external_types.PlanInstructionDiscriminator.REACTION_WINDOW_OPEN
                    | external_types.PlanInstructionDiscriminator.REACTION_WINDOW_CLOSE
                    | external_types.PlanInstructionDiscriminator.SUCCEED
                    | external_types.PlanInstructionDiscriminator.WAIT_END
                    | external_types.PlanInstructionDiscriminator.WAIT_START
                ):
                    final_tape.append(instruction)
                case external_types.PlanInstructionDiscriminator.JUMP:
                    instruction: internal_types.IntermediatePlanInstructionJump
                    target_offset = instruction["_target_offset"]
                    resolved_target = instruction_address + target_offset
                    jump_instruction = external_types.PlanInstructionJump(
                        type=external_types.PlanInstructionDiscriminator.JUMP,
                        target=resolved_target
                    )
                    final_tape.append(jump_instruction)
                case external_types.PlanInstructionDiscriminator.JUMP_IF_FALSE:
                    instruction: internal_types.IntermediatePlanInstructionJumpIfFalse
                    target_offset = instruction["_target_offset"]
                    resolved_target = instruction_address + target_offset
                    jump_if_false_instruction = external_types.PlanInstructionJumpIfFalse(
                        type=external_types.PlanInstructionDiscriminator.JUMP_IF_FALSE,
                        condition=instruction["_condition"],
                        target=resolved_target
                    )
                    final_tape.append(jump_if_false_instruction)
                case external_types.PlanInstructionDiscriminator.LOOP_NEXT:
                    instruction: internal_types.IntermediatePlanInstructionLoopNext
                    exit_target_offset = instruction["_exitTarget_offset"]
                    resolved_exit_target = instruction_address + exit_target_offset
                    loop_next_instruction = external_types.PlanInstructionLoopNext(
                        type=external_types.PlanInstructionDiscriminator.LOOP_NEXT,
                        exitTarget=resolved_exit_target
                    )
                    final_tape.append(loop_next_instruction)
                case _:
                    error_message = f"Unexpected plan-instruction type: {instruction['type']}"
                    raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        # Return the prepared tape
        return final_tape

    @staticmethod
    def visit_plan_instruction(
        _, children: list[list[internal_types.IntermediatePlanInstruction]]
    ) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_instruction> node.

        Note: Despite the singular name for this nonterminal, we always return a list of instructions
        here. Because some instruction types compile to lists of instructions (e.g., reaction windows),
        we proactively list-ify all instruction types as they are initially emitted by the dedicated
        visitor methods below. As such, this method always returns a list of intermediate instructions.
        """
        return children[0]

    @staticmethod
    def visit_plan_loop(_, children: list[Any]) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_loop> node."""
        iterable, variable, body_instructions = children
        loop_init_instruction = external_types.PlanInstructionLoopInit(
            type=external_types.PlanInstructionDiscriminator.LOOP_INIT,
            iterable=iterable,
            variable=variable
        )
        intermediate_loop_next_instruction = internal_types.IntermediatePlanInstructionLoopNext(
            type=external_types.PlanInstructionDiscriminator.LOOP_NEXT,
            # Jump forward past this instruction, the body instructions, and then past the trailing
            # jump. If it feels like there's an extra `+1`, it's because we always need to move by
            # at least one instruction to avoid staying put (i.e., a `+0`).
            _exitTarget_offset=(1 + len(body_instructions) + 1)
        )
        intermediate_jump_instruction = internal_types.IntermediatePlanInstructionJump(
            type=external_types.PlanInstructionDiscriminator.JUMP,
            # Jump backward past this instruction and the body instructions, all the way back to the
            # loop-next instruction. Again, if the `-1` feels strange here, it's because we always
            # need to move by at least one instruction to avoid staying put (i.e., a `-0`).
            _target_offset=(-1 - len(body_instructions))
        )
        instructions = (
            [loop_init_instruction, intermediate_loop_next_instruction] +
            body_instructions +
            [intermediate_jump_instruction]
        )
        return instructions

    @staticmethod
    def visit_plan_conditional(
        _, children: list[list[internal_types.IntermediatePlanConditionalBranch]]
    ) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_conditional> node."""
        # Count up the total number of instructions, since all branches need an offset
        # to jump to the end, past the other branches.
        branches: list[internal_types.IntermediatePlanConditionalBranch] = []
        for branches_group in children:
            branches += branches_group
        n_conditional_instructions = 0
        for branch in branches:
            if branch["_condition"]:
                n_conditional_instructions += 2  # Two jumps per branch
            n_conditional_instructions += len(branch["_consequent"])
        # Linearize the instructions that make up the plan conditional
        conditional_instructions_so_far: list[internal_types.IntermediatePlanInstruction] = []
        for branch in branches:
            branch_instructions: list[internal_types.IntermediatePlanInstruction] = branch["_consequent"]
            # If this is the alternative, simply append its instructions now
            if not branch["_condition"]:
                conditional_instructions_so_far += branch_instructions
                continue
            # Otherwise, it's a conditional branch, so we need to augment the instructions with jump machinery
            intermediate_jump_if_false_instruction = internal_types.IntermediatePlanInstructionJumpIfFalse(
                type=external_types.PlanInstructionDiscriminator.JUMP_IF_FALSE,
                _condition=branch["_condition"],
                # Jump forward past this instruction, the branch instructions, and then past the trailing
                # jump. If it feels like there's an extra `+1`, it's because we always need to move by
                # at least one instruction to avoid staying put (i.e., a `+0`).
                _target_offset=(1 + len(branch_instructions) + 1)
            )
            conditional_instructions_so_far += [intermediate_jump_if_false_instruction] + branch_instructions
            intermediate_jump_instruction = internal_types.IntermediatePlanInstructionJump(
                type=external_types.PlanInstructionDiscriminator.JUMP,
                # Jump forward to the first instruction after the conditional. This entails jumping past this
                # instruction (hence the leading `+1`) and then past all remaining branches and the alternative,
                # as applicable. This works because the number of trailing instructions in the conditional will
                # always be precisely `n_conditional_instructions - len(conditional_instructions_so_far)`.
                _target_offset=(1 + n_conditional_instructions - len(conditional_instructions_so_far))
            )
            conditional_instructions_so_far += [intermediate_jump_instruction]
        return conditional_instructions_so_far

    @staticmethod
    def visit_plan_conditional_branches(
        _, children: list[internal_types.IntermediatePlanConditionalBranch]
    ) -> list[internal_types.IntermediatePlanConditionalBranch]:
        """Visit a <plan_conditional_branches> node."""
        return children

    @staticmethod
    def visit_plan_conditional_branch(_, children: list[Any]) -> internal_types.IntermediatePlanConditionalBranch:
        """Visit a <plan_conditional_branch> node."""
        condition, consequent = children
        conditional_branch = internal_types.IntermediatePlanConditionalBranch(
            _condition=condition,
            _consequent=consequent
        )
        return conditional_branch

    @staticmethod
    def visit_plan_conditional_consequent(
        _, children: list[list[internal_types.IntermediatePlanInstruction]]
    ) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_conditional_consequent> node."""
        return children[0]

    @staticmethod
    def visit_plan_conditional_alternative(
        _, children: list[list[internal_types.IntermediatePlanInstruction]]
    ) -> list[internal_types.IntermediatePlanConditionalBranch]:
        """Visit a <plan_conditional_alternative> node."""
        conditional_alternative_branch = internal_types.IntermediatePlanConditionalBranch(
            _condition=None,
            _consequent=children[0]
        )
        # We'll return a list here for parity with `visit_plan_conditional_branches()`, which will
        # make it easier to linearize the conditional in `visit_plan_conditional()` up above.
        return [conditional_alternative_branch]

    @staticmethod
    def visit_plan_scoped_statements(
        _, children: list[list[internal_types.IntermediatePlanInstruction]]
    ) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_scoped_statements> node."""
        instructions: list[internal_types.IntermediatePlanInstruction] = []
        for instructions_group in children:
            instructions += instructions_group
        return instructions

    @staticmethod
    def visit_plan_instruction_reaction(_, children: list[Any]) -> list[external_types.PlanInstructionReactionQueue]:
        """Visit a <plan_instruction_reaction> node."""
        reaction_instruction = external_types.PlanInstructionReactionQueue(
            type=external_types.PlanInstructionDiscriminator.REACTION_QUEUE,
            reaction=children[0]
        )
        return [reaction_instruction]

    @staticmethod
    def visit_plan_instruction_reaction_window(
        node: NonTerminal,
        children: list[internal_types.Sentinel | list[internal_types.IntermediatePlanInstruction]]
    ) -> list[internal_types.IntermediatePlanInstruction]:
        """Visit a <plan_instruction_reaction_window> node."""
        operator_sentinel, *window_body_instruction_groups = children
        window_body_instructions: list[internal_types.IntermediatePlanInstruction] = []
        for instructions_group in window_body_instruction_groups:
            window_body_instructions += instructions_group
        if operator_sentinel is sentinels.REACTION_WINDOW_OPERATOR_ALL:
            operator = external_types.PlanPhaseReactionWindowOperator.ALL
        elif operator_sentinel is sentinels.REACTION_WINDOW_OPERATOR_ANY:
            operator = external_types.PlanPhaseReactionWindowOperator.ANY
        elif operator_sentinel is sentinels.REACTION_WINDOW_OPERATOR_UNTRACKED:
            # An `untracked` window is compiled as a sequence of bare reactions, since that's its
            # semantics and there's no need to spin up unused window machinery.
            return window_body_instructions
        else:
            raise errors.VivCompileError(
                "Unexpected operator for reaction window in plan phase",
                source=utils.derive_source_annotations(node=node)
            )
        window_open_instruction = external_types.PlanInstructionReactionWindowOpen(
            type=external_types.PlanInstructionDiscriminator.REACTION_WINDOW_OPEN
        )
        window_close_instruction = external_types.PlanInstructionReactionWindowClose(
            type=external_types.PlanInstructionDiscriminator.REACTION_WINDOW_CLOSE,
            operator=operator
        )
        instructions = [window_open_instruction] + window_body_instructions + [window_close_instruction]
        return instructions

    @staticmethod
    def visit_plan_instruction_reaction_window_operator(node: NonTerminal, _) -> internal_types.Sentinel:
        """Visit a <plan_instruction_reaction_window_operator> node."""
        match str(node):
            case "all":
                return sentinels.REACTION_WINDOW_OPERATOR_ALL
            case "any":
                return sentinels.REACTION_WINDOW_OPERATOR_ANY
            case "untracked":
                return sentinels.REACTION_WINDOW_OPERATOR_UNTRACKED
            case _:
                raise errors.VivCompileError(
                    "Unexpected operator for reaction window in plan phase",
                    source=utils.derive_source_annotations(node=node)
                )

    @staticmethod
    def visit_plan_instruction_wait(_, children: list[Any]) -> list[external_types.PlanInstruction]:
        """Visit a <plan_instruction_wait> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        wait_start_instruction = external_types.PlanInstructionWaitStart(
            type=external_types.PlanInstructionDiscriminator.WAIT_START,
            timeout=accumulated_fields["timeout"]
        )
        wait_end_instruction = external_types.PlanInstructionWaitEnd(
            type=external_types.PlanInstructionDiscriminator.WAIT_END,
            resumeConditions=accumulated_fields.get("resumeConditions", None)
        )
        return [wait_start_instruction, wait_end_instruction]

    @staticmethod
    def visit_plan_instruction_wait_timeout(
        _, children: list[external_types.TimeDelta]
    ) -> dict[str, external_types.TimeDelta]:
        """Visit a <plan_instruction_wait_timeout> node."""
        return {"timeout": children[0]}

    @staticmethod
    def visit_plan_instruction_wait_until(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <plan_instruction_wait_until> node."""
        return {"resumeConditions": children[0]}

    @staticmethod
    def visit_plan_instruction_advance(_, __) -> list[external_types.PlanInstructionAdvance]:
        """Visit a <plan_instruction_advance> node."""
        advance_instruction = external_types.PlanInstructionAdvance(
            type=external_types.PlanInstructionDiscriminator.ADVANCE
        )
        return [advance_instruction]

    @staticmethod
    def visit_plan_instruction_succeed(_, __) -> list[external_types.PlanInstructionSucceed]:
        """Visit a <plan_instruction_succeed> node."""
        succeed_instruction = external_types.PlanInstructionSucceed(
            type=external_types.PlanInstructionDiscriminator.SUCCEED
        )
        return [succeed_instruction]

    @staticmethod
    def visit_plan_instruction_fail(_, __) -> list[external_types.PlanInstructionFail]:
        """Visit a <plan_instruction_fail> node."""
        fail_instruction = external_types.PlanInstructionFail(
            type=external_types.PlanInstructionDiscriminator.FAIL
        )
        return [fail_instruction]

    @staticmethod
    def visit_plan_phase_sigil(_, __) -> internal_types.Sentinel:
        """Visit a <plan_phase_sigil> node."""
        return sentinels.PLAN_PHASE_SIGIL



"""Handles construction of the finalized compiled content bundle."""

__all__ = ["create_compiled_content_bundle"]

from viv_compiler import external_types, internal_types
from ._metadata import create_metadata


def create_compiled_content_bundle(combined_ast: internal_types.CombinedAST) -> external_types.ContentBundle:
    """Return a finalized compiled content bundle that is ready for validation.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A finalized compiled content bundle that is ready for validation.
    """
    # Create metadata to be attached to the compiled content bundle
    content_bundle_metadata = create_metadata(combined_ast=combined_ast)
    # Create the finalized construct definitions
    action_definitions = _create_action_definitions(combined_ast=combined_ast)
    action_selector_definitions = _create_action_selector_definitions(combined_ast=combined_ast)
    plan_definitions = _create_plan_definitions(combined_ast=combined_ast)
    plan_selector_definitions = _create_plan_selector_definitions(combined_ast=combined_ast)
    query_definitions = _create_query_definitions(combined_ast=combined_ast)
    sifting_pattern_definitions = _create_sifting_pattern_definitions(combined_ast=combined_ast)
    trope_definitions = _create_trope_definitions(combined_ast=combined_ast)
    # Package up and return the compiled content bundle
    compiled_content_bundle = external_types.ContentBundle(
        metadata=content_bundle_metadata,
        actions=action_definitions,
        actionSelectors=action_selector_definitions,
        plans=plan_definitions,
        queries=query_definitions,
        siftingPatterns=sifting_pattern_definitions,
        planSelectors=plan_selector_definitions,
        tropes=trope_definitions
    )
    return compiled_content_bundle


def _create_action_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.ActionName, external_types.ActionDefinition]:
    """Return a dictionary mapping action names to (finalized) action definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping action names to (finalized) action definitions.
    """
    action_definitions: dict[external_types.ActionName, external_types.ActionDefinition] = {}
    for intermediate_action_definition in combined_ast["actions"]:
        action_definition = external_types.ActionDefinition(
            type=intermediate_action_definition["type"],
            name=intermediate_action_definition["name"],
            reserved=intermediate_action_definition["reserved"],
            roles=intermediate_action_definition["roles"],
            initiator=intermediate_action_definition["initiator"],
            roleForestRoots=intermediate_action_definition["roleForestRoots"],
            importance=intermediate_action_definition["importance"],
            tags=intermediate_action_definition["tags"],
            gloss=intermediate_action_definition["gloss"],
            report=intermediate_action_definition["report"],
            conditions=intermediate_action_definition["conditions"],
            scratch=intermediate_action_definition["scratch"],
            effects=intermediate_action_definition["effects"],
            reactions=intermediate_action_definition["reactions"],
            saliences=intermediate_action_definition["saliences"],
            associations=intermediate_action_definition["associations"],
            embargoes=intermediate_action_definition["embargoes"],
        )
        action_definitions[intermediate_action_definition["name"]] = action_definition
    return action_definitions


def _create_action_selector_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.SelectorName, external_types.ActionSelectorDefinition]:
    """Return a dictionary mapping action-selector names to (finalized) action-selector definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping action-selector names to (finalized) action-selector definitions.
    """
    action_selector_definitions: dict[external_types.SelectorName, external_types.ActionSelectorDefinition] = {}
    for intermediate_action_selector_definition in combined_ast["actionSelectors"]:
        action_selector_definition = external_types.ActionSelectorDefinition(
            type=intermediate_action_selector_definition["type"],
            name=intermediate_action_selector_definition["name"],
            reserved=intermediate_action_selector_definition["reserved"],
            roles=intermediate_action_selector_definition["roles"],
            initiator=intermediate_action_selector_definition["initiator"],
            roleForestRoots=intermediate_action_selector_definition["roleForestRoots"],
            conditions=intermediate_action_selector_definition["conditions"],
            policy=intermediate_action_selector_definition["policy"],
            candidates=intermediate_action_selector_definition["candidates"]
        )
        action_selector_definitions[intermediate_action_selector_definition["name"]] = action_selector_definition
    return action_selector_definitions


def _create_plan_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.PlanName, external_types.PlanDefinition]:
    """Return a dictionary mapping plan names to (finalized) plan definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping plan names to (finalized) plan definitions.
    """
    plan_definitions: dict[external_types.PlanName, external_types.PlanDefinition] = {}
    for intermediate_plan_definition in combined_ast["plans"]:
        plan_definition = external_types.PlanDefinition(
            type=intermediate_plan_definition["type"],
            name=intermediate_plan_definition["name"],
            roles=intermediate_plan_definition["roles"],
            roleForestRoots=intermediate_plan_definition["roleForestRoots"],
            conditions=intermediate_plan_definition["conditions"],
            phases=intermediate_plan_definition["phases"],
            initialPhase=intermediate_plan_definition["initialPhase"]
        )
        plan_definitions[intermediate_plan_definition["name"]] = plan_definition
    return plan_definitions


def _create_plan_selector_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.SelectorName, external_types.PlanSelectorDefinition]:
    """Return a dictionary mapping plan-selector names to (finalized) plan-selector definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping plan-selector names to (finalized) plan-selector definitions.
    """
    plan_selector_definitions: dict[external_types.SelectorName, external_types.PlanSelectorDefinition] = {}
    for intermediate_plan_selector_definition in combined_ast["planSelectors"]:
        plan_selector_definition = external_types.PlanSelectorDefinition(
            type=intermediate_plan_selector_definition["type"],
            name=intermediate_plan_selector_definition["name"],
            roles=intermediate_plan_selector_definition["roles"],
            roleForestRoots=intermediate_plan_selector_definition["roleForestRoots"],
            conditions=intermediate_plan_selector_definition["conditions"],
            policy=intermediate_plan_selector_definition["policy"],
            candidates=intermediate_plan_selector_definition["candidates"]
        )
        plan_selector_definitions[intermediate_plan_selector_definition["name"]] = plan_selector_definition
    return plan_selector_definitions


def _create_query_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.QueryName, external_types.QueryDefinition]:
    """Return a dictionary mapping query names to (finalized) query definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping query names to (finalized) query definitions.
    """
    query_definitions: dict[external_types.QueryName, external_types.QueryDefinition] = {}
    for intermediate_query_definition in combined_ast["queries"]:
        query_definition = external_types.QueryDefinition(
            type=intermediate_query_definition["type"],
            name=intermediate_query_definition["name"],
            roles=intermediate_query_definition["roles"],
            roleForestRoots=intermediate_query_definition["roleForestRoots"],
            conditions=intermediate_query_definition["conditions"],
            actionName=intermediate_query_definition["actionName"],
            ancestors=intermediate_query_definition["ancestors"],
            descendants=intermediate_query_definition["descendants"],
            importance=intermediate_query_definition["importance"],
            tags=intermediate_query_definition["tags"],
            salience=intermediate_query_definition["salience"],
            associations=intermediate_query_definition["associations"],
            location=intermediate_query_definition["location"],
            time=intermediate_query_definition["time"],
            initiator=intermediate_query_definition["initiator"],
            partners=intermediate_query_definition["partners"],
            recipients=intermediate_query_definition["recipients"],
            bystanders=intermediate_query_definition["bystanders"],
            active=intermediate_query_definition["active"],
            present=intermediate_query_definition["present"]
        )
        query_definitions[intermediate_query_definition["name"]] = query_definition
    return query_definitions


def _create_sifting_pattern_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.SiftingPatternName, external_types.SiftingPatternDefinition]:
    """Return a dictionary mapping sifting-pattern names to (finalized) sifting-pattern definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping sifting-pattern names to (finalized) sifting-pattern definitions.
    """
    sifting_pattern_definitions: dict[
        external_types.SiftingPatternName,
        external_types.SiftingPatternDefinition,
    ] = {}
    for intermediate_sifting_pattern_definition in combined_ast["siftingPatterns"]:
        sifting_pattern_definition = external_types.SiftingPatternDefinition(
            type=intermediate_sifting_pattern_definition["type"],
            name=intermediate_sifting_pattern_definition["name"],
            roles=intermediate_sifting_pattern_definition["roles"],
            roleForestRoots=intermediate_sifting_pattern_definition["roleForestRoots"],
            conditions=intermediate_sifting_pattern_definition["conditions"],
            actions=intermediate_sifting_pattern_definition["actions"]
        )
        sifting_pattern_definitions[intermediate_sifting_pattern_definition["name"]] = sifting_pattern_definition
    return sifting_pattern_definitions


def _create_trope_definitions(
    combined_ast: internal_types.CombinedAST
) -> dict[external_types.TropeName, external_types.TropeDefinition]:
    """Return a dictionary mapping trope names to (finalized) trope definitions.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A dictionary mapping trope names to (finalized) trope definitions.
    """
    trope_definitions: dict[external_types.TropeName, external_types.TropeDefinition] = {}
    for intermediate_trope_definition in combined_ast["tropes"]:
        trope_definition = external_types.TropeDefinition(
            type=intermediate_trope_definition["type"],
            name=intermediate_trope_definition["name"],
            roles=intermediate_trope_definition["roles"],
            roleForestRoots=intermediate_trope_definition["roleForestRoots"],
            conditions=intermediate_trope_definition["conditions"]
        )
        trope_definitions[intermediate_trope_definition["name"]] = trope_definition
    return trope_definitions



"""Internal functionality for validating plan definitions."""

__all__ = ["validate_plan_definitions"]

from viv_compiler import errors, external_types
from ._common import validate_common_concerns


def validate_plan_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the plan definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    for plan_definition in content_bundle["plans"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=plan_definition, content_bundle=content_bundle)
        # Validate reaction windows
        _validate_plan_reaction_windows(plan_definition=plan_definition)


def _validate_plan_reaction_windows(plan_definition: external_types.PlanDefinition) -> None:
    """Validate the reaction windows in the given plan definition.

    Args:
        plan_definition: A plan definition from a compiled content bundle.

    Returns:
        None. Only returns if no issue was detected.

    Raises:
        VivCompileError: The plan definition has overlapping or nested reaction windows.
    """
    for phase_name, phase_definition in plan_definition['phases'].items():
        reaction_window_depth = 0
        for instruction in phase_definition['tape']:
            if instruction['type'] == external_types.PlanInstructionDiscriminator.REACTION_WINDOW_OPEN:
                reaction_window_depth += 1
                if reaction_window_depth > 1:
                    raise errors.VivCompileError(
                        f"Plan '{plan_definition['name']}' has nested/overlapping reaction windows in phase "
                        f"'{phase_name}' (reaction windows must be used sequentially, with no nesting/overlap)"
                    )
            elif instruction['type'] == external_types.PlanInstructionDiscriminator.REACTION_WINDOW_CLOSE:
                reaction_window_depth -= 1



"""API for the Viv compiler."""

__all__ = ["compile_from_path"]

from pathlib import Path

from viv_compiler import config, pipeline, external_types


def compile_from_path(
    *,  # Require keyword arguments only
    source_file_path: Path,
    default_importance: float = config.DEFAULT_IMPORTANCE_SCORE,
    default_salience: float = config.DEFAULT_SALIENCE_SCORE,
    default_reaction_priority: float = config.DEFAULT_REACTION_PRIORITY_VALUE,
    use_memoization: bool = True,
    verbose_parser: bool = False
) -> external_types.ContentBundle:
    """Compile the given Viv source file to produce a JSON-serializable compiled content bundle.

    Args:
        source_file_path: Relative or absolute path to the Viv source file to be parsed.
        default_importance: A default importance value to use when one is not specified
            in an action definition.
        default_salience: A default salience value to use when one is not specified
            in an action definition.
        default_reaction_priority: A default reaction priority to use when one is not
            specified in a reaction declaration.
        use_memoization: Whether to use memoization during PEG parsing (faster, but uses more memory).
        verbose_parser: Whether to invoke verbose debugging for the PEG parser itself.

    Returns:
        The compiled content bundle.

    Raises:
        VivParseError: The source file, or a file it includes, could not be parsed.
        VivCompileError: An issue occurred during compilation, but after parsing.
    """
    return pipeline.compile_viv_source_code(
        source_file_path=source_file_path,
        default_importance=default_importance,
        default_salience=default_salience,
        default_reaction_priority=default_reaction_priority,
        use_memoization=use_memoization,
        verbose_parser=verbose_parser,
    )



"""Module that orchestrates the core pipeline for the Viv compiler.

This high-level module invokes other components to orchestrate the full compilation pipeline:

  * Load and parse the Viv DSL grammar to build a Viv PEG parser.
  * Load and parse an entry Viv source file.
  * Walk the parse tree with a visitor, to produce an AST.
  * Honor `include` statements (imports between Viv files), by producing and combining ASTs.
  * Postprocess the combined AST to produce finalized construct definitions.
  * Package up the construct definitions into a proper compiled content bundle.
  * Validate the content bundle, including against a canonical schema.
  * Emit JSON-serializable output for the validated content bundle.
"""

from .pipeline import compile_viv_source_code

__all__ = ["compile_viv_source_code"]



"""Mixin class containing visitor methods associated with Viv expressions."""

__all__ = ["VisitorMixinExpressions"]

from typing import Any, cast

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import config, internal_types, errors, external_types, sentinels, utils


class VisitorMixinExpressions(PTNodeVisitor):
    """A visitor mixin for Viv expressions."""

    @staticmethod
    def visit_expression(_, children: list[Any]) -> external_types.Expression:
        """Visit an <expression> node."""
        return children[0]

    @staticmethod
    def visit_negation(_, __) -> internal_types.Sentinel:
        """Visit a <negation> node."""
        return sentinels.NEGATION

    @staticmethod
    def visit_assignment(node: NonTerminal, children: list[Any]) -> external_types.Assignment:
        """Visit an <assignment> node."""
        left, operator, right = children
        assignment = external_types.Assignment(
            type=external_types.ExpressionDiscriminator.ASSIGNMENT,
            value=external_types.AssignmentValue(left=left, operator=operator, right=right),
            source=utils.derive_source_annotations(node=node)
        )
        return assignment

    @staticmethod
    def visit_assignment_operator(node: NonTerminal, _) -> str:
        """Visit an <assignment_operator> node."""
        return str(node)

    @staticmethod
    def visit_assignment_lvalue(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <assignment_lvalue> node."""
        return children[0]

    @staticmethod
    def visit_inscription(node: NonTerminal, children: list[Any]) -> external_types.Inscription:
        """Visit an <inscription> node."""
        item, action = children
        inscription = external_types.Inscription(
            type=external_types.ExpressionDiscriminator.INSCRIPTION,
            value=external_types.InscriptionValue(item=item, action=action),
            source=utils.derive_source_annotations(node=node)
        )
        return inscription

    @staticmethod
    def visit_inspection(node: NonTerminal, children: list[Any]) -> external_types.Inspection:
        """Visit an <inspection> node."""
        character, item = children
        inspection = external_types.Inspection(
            type=external_types.ExpressionDiscriminator.INSPECTION,
            value=external_types.InspectionValue(character=character, item=item),
            source=utils.derive_source_annotations(node=node)
        )
        return inspection

    @staticmethod
    def visit_precedence_governed_expression(_, children: list[Any]) -> external_types.Expression:
        """Visit a <precedence_governed_expression> node."""
        # The <precedence_governed_expression> nonterminal is just a wrapper around disjunction. All the work
        # to package up an expression -- potentially one that is quite complex and mixes operators without using
        # parens -- will have already been done by the time it gets to here, so we simply return the sole child.
        return children[0]

    @staticmethod
    def visit_disjunction(node: NonTerminal, children: list[Any]) -> external_types.Disjunction:
        """Visit a <disjunction> node."""
        # If there's a single child, it's an expression with no disjunction operator bubbling up
        # through the precedence chain. As such, we want to just pass it on up as-is.
        if len(children) == 1:
            return children[0]
        disjunction = external_types.Disjunction(
            type=external_types.ExpressionDiscriminator.DISJUNCTION,
            value=external_types.DisjunctionValue(operands=children),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )
        return disjunction

    @staticmethod
    def visit_conjunction(node: NonTerminal, children: list[Any]) -> external_types.Conjunction:
        """Visit a <conjunction> node."""
        # If there's a single child, it's an expression with no conjunction operator bubbling up
        # through the precedence chain. As such, we want to just pass it on up as-is.
        if len(children) == 1:
            return children[0]
        conjunction = external_types.Conjunction(
            type=external_types.ExpressionDiscriminator.CONJUNCTION,
            value=external_types.ConjunctionValue(operands=children),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )
        return conjunction

    @staticmethod
    def visit_relational_expression(node: NonTerminal, children: list[Any]) -> external_types.Expression:
        """Visit a <relational_expression> node."""
        # If there's a single child, it's an expression with no relational operator bubbling up
        # through the precedence chain. As such, we want to just pass it on up as-is.
        if len(children) == 1:
            return children[0]
        left, operator, right = children
        if operator == "in":
            return external_types.MembershipTest(
                type=external_types.ExpressionDiscriminator.MEMBERSHIP_TEST,
                value=external_types.MembershipTestValue(item=left, collection=right),
                source=utils.derive_source_annotations(node=node),
                negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
            )
        elif operator == "knows":
            return external_types.MemoryCheck(
                type=external_types.ExpressionDiscriminator.MEMORY_CHECK,
                value=external_types.MemoryCheckValue(character=left, action=right),
                source=utils.derive_source_annotations(node=node),
                negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
            )
        elif operator in ("caused", "triggered", "preceded"):
            return external_types.ActionRelation(
                type=external_types.ExpressionDiscriminator.ACTION_RELATION,
                value=external_types.ActionRelationValue(left=left, operator=operator, right=right),
                source=utils.derive_source_annotations(node=node),
                negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
            )
        return external_types.Comparison(
            type=external_types.ExpressionDiscriminator.COMPARISON,
            value=external_types.ComparisonValue(left=left, operator=operator, right=right),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )

    @staticmethod
    def visit_relational_operator(node: NonTerminal, _) -> str:
        """Visit a <relational_operator> node."""
        return str(node)

    @staticmethod
    def visit_additive_expression(node: NonTerminal, children: list[Any]) -> external_types.Expression:
        """Visit an <additive_expression> node."""
        # If there's a single child, it's an expression with no additive operator bubbling up
        # through the precedence chain. As such, we want to just pass it on up as-is.
        if len(children) == 1:
            return children[0]
        result = children[0]
        for i in range(1, len(children), 2):  # Children arrive as [term, op, term, op, term, ...]
            operator = cast(external_types.ArithmeticOperator, str(children[i]))
            right = children[i + 1]
            result = external_types.ArithmeticExpression(
                type=external_types.ExpressionDiscriminator.ARITHMETIC_EXPRESSION,
                value=external_types.ArithmeticExpressionValue(left=result, operator=operator, right=right),
                source=utils.derive_source_annotations(node=node)
            )
        return result

    @staticmethod
    def visit_additive_operator(node: NonTerminal, _) -> str:
        """Visit a <additive_operator> node."""
        return str(node)

    @staticmethod
    def visit_multiplicative_expression(node: NonTerminal, children: list[Any]) -> external_types.Expression:
        """Visit a <multiplicative_expression> node."""
        # If there's a single child, it's an expression with no multiplicative operator bubbling up
        # through the precedence chain. As such, we want to just pass it on up as-is.
        if len(children) == 1:
            return children[0]
        return VisitorMixinExpressions.visit_additive_expression(node=node, children=children)

    @staticmethod
    def visit_multiplicative_operator(node: NonTerminal, _) -> str:
        """Visit a <multiplicative_operator> node."""
        return str(node)

    @staticmethod
    def visit_unary_expression(_, children: list[Any]) -> external_types.Expression:
        """Visit a <unary_expression> node."""
        unary_expression = children[-1]
        # If applicable, mutate the unary expression to mark it as being negated. The only way for an expression
        # to become negated is right here, since in the Viv grammar the negation operator can only appear inside
        # a `unary_expression`. To negate a complex expression such as a conjunction, the author must wrap it in
        # parentheses, which ultimately causes it to be parsed as a unary expression.
        if children[0] is sentinels.NEGATION:
            unary_expression['negated'] = True
        return unary_expression

    @staticmethod
    def visit_reference(
        node: NonTerminal,
        children: list[Any]
    ) -> external_types.EntityReference | external_types.SymbolReference:
        """Visit a <reference> node."""
        is_symbol_reference = False
        anchor_is_local_variable = False
        anchor_is_scratch_variable = False
        anchor_is_group_role = False
        anchor_fail_safe = False
        anchor_name = None
        path = []
        for child in children:
            if child is sentinels.ENTITY_SIGIL:
                pass
            elif child is sentinels.SYMBOL_SIGIL:
                is_symbol_reference = True
            elif child is sentinels.LOCAL_VARIABLE_SIGIL:
                anchor_is_local_variable = True
            elif child is sentinels.SCRATCH_VARIABLE_SIGIL:
                anchor_is_scratch_variable = True
            elif child is sentinels.GROUP_ROLE_DECORATOR:
                anchor_is_group_role = True
            elif child is sentinels.EVAL_FAIL_SAFE_MARKER:
                anchor_fail_safe = True
            elif type(child) is str:
                anchor_name = child
            else:
                path = child
        if anchor_is_scratch_variable:
            # We will ignore the symbol sigil here, because ultimately that refers to the type of
            # the scratch variable, not the anchor, which is always entity data.
            is_symbol_reference = False
            # Next, expand the `$` anchor. `$` is really just syntactic sugar for `@this.scratch.`,
            # which means any reference anchored in a scratch variable is in fact an entity
            # reference anchored in a role name.
            original_anchor_name = anchor_name
            anchor_name = config.SCRATCH_VARIABLE_REFERENCE_ANCHOR
            expanded_path = [
                *config.SCRATCH_VARIABLE_REFERENCE_PATH_PREFIX,
                external_types.ReferencePathComponentPropertyName(
                    type=external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME,
                    name=original_anchor_name,
                    failSafe=anchor_fail_safe
                )
            ]
            path = expanded_path + path
            anchor_fail_safe = False
        reference_value = external_types.ReferenceValue(
            anchor=anchor_name,
            path=path,
            local=anchor_is_local_variable,
            group=anchor_is_group_role,
            failSafe=anchor_fail_safe
        )
        if is_symbol_reference:
            return external_types.SymbolReference(
                type=external_types.ExpressionDiscriminator.SYMBOL_REFERENCE,
                value=reference_value,
                source=utils.derive_source_annotations(node=node),
                negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
            )
        return external_types.EntityReference(
            type=external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
            value=reference_value,
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )

    @staticmethod
    def visit_reference_path(_, children: list[Any]) -> list[external_types.ReferencePathComponent]:
        """Visit a <reference_path> node."""
        return children

    @staticmethod
    def visit_reference_path_property_name(_, children: list[Any]) -> external_types.ReferencePathComponentPropertyName:
        """Visit a <reference_path_property_name> node."""
        property_name = children[0]
        reference_path_component_property_name = external_types.ReferencePathComponentPropertyName(
            type=external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME,
            name=property_name,
            failSafe=len(children) > 1
        )
        return reference_path_component_property_name

    @staticmethod
    def visit_reference_path_pointer(_, children: list[Any]) -> external_types.ReferencePathComponentPointer:
        """Visit a <reference_path_pointer> node."""
        property_name = children[0]
        reference_path_component_pointer = external_types.ReferencePathComponentPointer(
            type=external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_POINTER,
            propertyName=property_name,
            failSafe=len(children) > 1
        )
        return reference_path_component_pointer

    @staticmethod
    def visit_reference_path_lookup(_, children: list[Any]) -> external_types.ReferencePathComponentLookup:
        """Visit a <reference_path_lookup> node."""
        key = children[0]
        reference_path_component_lookup = external_types.ReferencePathComponentLookup(
            type=external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_LOOKUP,
            key=key,
            failSafe=len(children) > 1
        )
        return reference_path_component_lookup

    @staticmethod
    def visit_list(node: NonTerminal, children: list[Any]) -> external_types.ListField:
        """Visit a <list> node."""
        list_field = external_types.ListField(
            type=external_types.ExpressionDiscriminator.LIST,
            value=children,
            source=utils.derive_source_annotations(node=node)
        )
        return list_field

    @staticmethod
    def visit_object(node: NonTerminal, children: list[Any]) -> external_types.ObjectField:
        """Visit an <object> node."""
        object_field_value = {}
        for child in children:
            object_field_value.update(child)
        object_field = external_types.ObjectField(
            type=external_types.ExpressionDiscriminator.OBJECT,
            value=object_field_value,
            source=utils.derive_source_annotations(node=node)
        )
        return object_field

    @staticmethod
    def visit_key_value_pair(_, children: list[Any]) -> dict[str, external_types.Expression]:
        """Visit a <key_value_pair> node."""
        key, value = children
        if isinstance(key, dict):  # The author formatted the key as a string
            key = key['value']
        key_value_pair = {key: value}
        return key_value_pair

    @staticmethod
    def visit_custom_function_call(node: NonTerminal, children: list[Any]) -> external_types.CustomFunctionCall:
        """Visit a <custom_function_call> node."""
        function_result_fail_safe = False
        if children[-1] is sentinels.EVAL_FAIL_SAFE_MARKER:
            function_result_fail_safe = True
            children = children[:-1]
        if len(children) >= 2:
            function_name, function_args = children[0], children[1]
        else:  # No arguments passed in the function call, which is syntactically fine
            function_name = children[0]
            function_args = []
        custom_function_call = external_types.CustomFunctionCall(
            type=external_types.ExpressionDiscriminator.CUSTOM_FUNCTION_CALL,
            value=external_types.CustomFunctionCallValue(
                name=function_name,
                args=function_args,
                failSafe=function_result_fail_safe
            ),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )
        return custom_function_call

    @staticmethod
    def visit_args(_, children: list[Any]) -> list[external_types.Expression]:
        """Visit an <args> node."""
        return children

    @staticmethod
    def visit_chance_expression(node: NonTerminal, children: list[Any]) -> external_types.ChanceExpression:
        """Visit a <chance_expression> node."""
        chance = children[0]['value']
        probability = chance / 100
        chance_expression = external_types.ChanceExpression(
            type=external_types.ExpressionDiscriminator.CHANCE_EXPRESSION,
            value=probability,
            source=utils.derive_source_annotations(node=node)
        )
        return chance_expression

    @staticmethod
    def visit_literal(_, children: list[Any]) -> external_types.Expression:
        """Visit a <literal> node."""
        return children[0]

    @staticmethod
    def visit_enum(node: NonTerminal, children: list[Any]) -> external_types.Enum:
        """Visit an <enum> node."""
        additive_inverse_is_present = False
        if len(children) > 1:
            additive_inverse_is_present = children[0] == "-"
        name = children[-1]
        enum_expression = external_types.Enum(
            type=external_types.ExpressionDiscriminator.ENUM,
            value=external_types.EnumValue(name=name, minus=additive_inverse_is_present),
            source=utils.derive_source_annotations(node=node)
        )
        return enum_expression

    @staticmethod
    def visit_string(_, children: list[Any]) -> external_types.TemplateStringField | external_types.StringField:
        """Visit a <string> node."""
        return children[0]

    @staticmethod
    def visit_string_literal(node: NonTerminal, children: list[Any]) -> external_types.StringField:
        """Visit a <string_literal> node."""
        string_field = external_types.StringField(
            type=external_types.ExpressionDiscriminator.STRING,
            value=children[0],
            source=utils.derive_source_annotations(node=node)
        )
        return string_field

    @staticmethod
    def visit_template_string(node: NonTerminal, children: list[Any]) -> external_types.TemplateStringField:
        """Visit a <template_string> node."""
        # Because Arpeggio skips whitespace by default, a relatively elaborate procedure is required here
        # to maintain any whitespace embedded in the static portion of the template string. We'll do this
        # by walking the raw nodes associated with the children of this one, handling both gaps and static
        # elements. For gaps, we emit the corresponding visited expression. Between a gap and an adjacent
        # static element, we check source positions to detect whitespace that the parser gobbled, which we
        # then carefully re-inject into the template string.
        child_nodes = list(node)
        gap_expressions = [child for child in children if not isinstance(child, str)]
        template = []
        partial_string = ""
        gap_index = 0
        for i, raw_child in enumerate(child_nodes):
            child_text = raw_child.flat_str()
            # Skip enclosing quotes
            if child_text in ('"', "'"):
                continue
            if raw_child.rule_name == 'template_gap':
                # Flush accumulated text
                if partial_string:
                    template.append(partial_string)
                    partial_string = ""
                # Emit the visited expression for this gap
                template.append(gap_expressions[gap_index])
                gap_index += 1
                # If the parser gobbled whitespace after this gap, re-inject it in our
                # seed the next text segment.
                gap_size = child_nodes[i+1].position - raw_child.position_end
                if gap_size > 0:
                    partial_string = " " * gap_size
            else:
                # Accumulate text from this terminal
                partial_string += child_text
        # Flush any remaining text before constructing and returning the expression
        if partial_string:
            template.append(partial_string)
        template_string_field = external_types.TemplateStringField(
            type=external_types.ExpressionDiscriminator.TEMPLATE_STRING,
            value=template,
            source=utils.derive_source_annotations(node=node)
        )
        return template_string_field

    @staticmethod
    def visit_template_gap(_, children: list[Any]) -> external_types.Expression:
        """Visit a <template_gap> node."""
        return children[0]

    @staticmethod
    def visit_tags(node: NonTerminal, children: list[Any]) -> external_types.ListField:
        """Visit a <tags> node."""
        return VisitorMixinExpressions.visit_list(node=node, children=children)

    @staticmethod
    def visit_tag(node: NonTerminal, children: list[Any]) -> external_types.StringField:
        """Visit a <tag> node."""
        return VisitorMixinExpressions.visit_string_literal(node=node, children=children)

    @staticmethod
    def visit_number(node: NonTerminal, children: list[str]) -> external_types.FloatField | external_types.IntField:
        """Visit a <number> node."""
        number_str = ''.join(children)
        number = float(number_str)
        if number % 1:
            return external_types.FloatField(
                type=external_types.ExpressionDiscriminator.FLOAT,
                value=number,
                source=utils.derive_source_annotations(node=node)
            )
        return external_types.IntField(
            type=external_types.ExpressionDiscriminator.INT,
            value=int(number),
            source=utils.derive_source_annotations(node=node)
        )

    @staticmethod
    def visit_sign(node: NonTerminal, _) -> str:
        """Visit a <sign> node."""
        return str(node)

    @staticmethod
    def visit_boolean(node: NonTerminal, _) -> external_types.BoolField:
        """Visit a <boolean> node."""
        bool_field = external_types.BoolField(
            type=external_types.ExpressionDiscriminator.BOOL,
            value=(str(node) == "true"),
            source=utils.derive_source_annotations(node=node)
        )
        return bool_field

    @staticmethod
    def visit_null_type(node: NonTerminal, _) -> external_types.NullField:
        """Visit a <null_type> node."""
        null_field = external_types.NullField(
            type=external_types.ExpressionDiscriminator.NULL_TYPE,
            value=None,
            source=utils.derive_source_annotations(node=node)
        )
        return null_field

    @staticmethod
    def visit_trope_fit(node: NonTerminal, children: list[Any]) -> external_types.TropeFit:
        """Visit a <trope_fit> node."""
        trope_name, bindings_object = children
        bindings = bindings_object["bindings"]
        trope_fit = external_types.TropeFit(
            type=external_types.ExpressionDiscriminator.TROPE_FIT,
            value=external_types.TropeFitValue(tropeName=trope_name, bindings=bindings),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )
        return trope_fit

    @staticmethod
    def visit_trope_fit_sugared(node: NonTerminal, children: list[Any]) -> external_types.TropeFit:
        """Visit a <trope_fit_sugared> node.

        Note that there may be positional bindings in a sugared trope fit,
        which are resolved (to roles from indices) during postprocessing.
        """
        bindings_object, trope_name = children
        bindings = bindings_object["bindings"]
        trope_fit = external_types.TropeFit(
            type=external_types.ExpressionDiscriminator.TROPE_FIT,
            value=external_types.TropeFitValue(tropeName=trope_name, bindings=bindings),
            source=utils.derive_source_annotations(node=node),
            negated=False  # May be set to `True` by `visit_unary_expression()`, as applicable
        )
        return trope_fit

    @staticmethod
    def visit_action_search(node: NonTerminal, children: list[Any]) -> external_types.ActionSearch:
        """Visit an <action_search> node."""
        query_name, action_search_body = children
        if query_name is sentinels.ACTION_SEARCH_NO_QUERY:
            query_name = None
        if "searchDomain" not in action_search_body:
            error_message = "Action search is missing search domain ('over' field is required)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        action_search = external_types.ActionSearch(
            type=external_types.ExpressionDiscriminator.ACTION_SEARCH,
            value=external_types.ActionSearchValue(
                queryName=query_name,
                searchDomain=action_search_body["searchDomain"],
                bindings=action_search_body.get("bindings", external_types.PrecastBindings(partial=True, roles={}))
            ),
            source=utils.derive_source_annotations(node=node)
        )
        return action_search

    @staticmethod
    def visit_action_search_header(_, children: list[Any]) -> external_types.QueryName:
        """Visit an <action_search_header> node."""
        return children[0]

    @staticmethod
    def visit_action_search_bare_header(_, __) -> internal_types.Sentinel:
        """Visit an <action_search_bare_header> node."""
        return sentinels.ACTION_SEARCH_NO_QUERY

    @staticmethod
    def visit_action_search_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_search_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_action_search_bare_body(
            _, children: list[dict[str, external_types.SearchDomainDeclaration]]
    ) -> dict[str, external_types.SearchDomainDeclaration]:
        """Visit an <action_search_bare_body> node."""
        return children[0]

    @staticmethod
    def visit_search_domain(
        _, children: list[internal_types.Sentinel | external_types.Expression]
    ) -> dict[str, external_types.SearchDomainDeclaration]:
        """Visit a <search_domain> node."""
        raw_domain = children[0]
        domain_expression = None
        if raw_domain is sentinels.SEARCH_DOMAIN_INHERIT:
            domain_policy = external_types.SearchDomainPreparationPolicy.INHERIT
        elif raw_domain is sentinels.SEARCH_DOMAIN_CHRONICLE:
            domain_policy = external_types.SearchDomainPreparationPolicy.CHRONICLE
        else:
            domain_policy = external_types.SearchDomainPreparationPolicy.EXPRESSION
            domain_expression = raw_domain
        search_domain = external_types.SearchDomainDeclaration(
            policy=domain_policy,
            expression=domain_expression
        )
        return {"searchDomain": search_domain}

    @staticmethod
    def visit_search_domain_inherit(_, __) -> internal_types.Sentinel:
        """Visit a <search_domain_inherit> node."""
        return sentinels.SEARCH_DOMAIN_INHERIT

    @staticmethod
    def visit_search_domain_chronicle(_, __) -> internal_types.Sentinel:
        """Visit a <search_domain_chronicle> node."""
        return sentinels.SEARCH_DOMAIN_CHRONICLE

    @staticmethod
    def visit_search_domain_expression(_, children: list[external_types.Expression]) -> external_types.Expression:
        """Visit a <search_domain_expression> node."""
        return children[0]

    @staticmethod
    def visit_sifting(node: NonTerminal, children: list[Any]) -> external_types.Sifting:
        """Visit a <sifting> node."""
        pattern_name, sifting_body = children
        if "searchDomain" not in sifting_body:
            raise errors.VivCompileError(
                "Sifting expression is missing search domain (use 'over' field)",
                source=utils.derive_source_annotations(node=node)
            )
        sifting = external_types.Sifting(
            type=external_types.ExpressionDiscriminator.SIFTING,
            value=external_types.SiftingValue(
                patternName=pattern_name,
                searchDomain=sifting_body.get("searchDomain"),
                bindings=sifting_body.get("bindings", external_types.PrecastBindings(partial=True, roles={}))
            ),
            source=utils.derive_source_annotations(node=node)
        )
        return sifting

    @staticmethod
    def visit_sifting_header(_, children: list[Any]) -> external_types.SiftingPatternName:
        """Visit a <sifting_header> node."""
        return children[0]

    @staticmethod
    def visit_sifting_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <sifting_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_local_variable_sigil(_, __) -> internal_types.Sentinel:
        """Visit a <local_variable_sigil> node."""
        return sentinels.LOCAL_VARIABLE_SIGIL

    @staticmethod
    def visit_scratch_variable_sigil(_, __) -> internal_types.Sentinel:
        """Visit a <scratch_variable_sigil> node."""
        return sentinels.SCRATCH_VARIABLE_SIGIL

    @staticmethod
    def visit_group_role_decorator(_, __) -> internal_types.Sentinel:
        """Visit a <group_role_decorator> node."""
        return sentinels.GROUP_ROLE_DECORATOR

    @staticmethod
    def visit_eval_fail_safe_marker(_, __) -> internal_types.Sentinel:
        """Visit an <eval_fail_safe_marker> node."""
        return sentinels.EVAL_FAIL_SAFE_MARKER

    @staticmethod
    def visit_identifier(_, children: list[str]) -> str:
        """Visit an <identifier> node."""
        return ''.join(children)



"""Module serving as the visitor component of the Viv DSL compiler.

Exposes a single public visitor class that traverses and modifies a Viv AST, in accordance
with the structure prescribed by Arpeggio, our PEG parser solution. This class is composed
by a number of thematic mixin classes.

This implements the *visitor pattern* in parsing.
"""

from .visitor import Visitor



"""Internal functionality for validating action definitions."""

__all__ = ["validate_action_definitions"]

from viv_compiler import errors, external_types
from ._common import validate_common_concerns
from ._utils import get_all_action_role_names, is_non_numeric_expression


def validate_action_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the action definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    for action_definition in content_bundle["actions"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=action_definition, content_bundle=content_bundle)
        # Validate saliences
        _validate_action_saliences(action_definition=action_definition)
        # Validate associations
        _validate_action_associations(action_definition=action_definition)


def _validate_action_saliences(action_definition: external_types.ActionDefinition) -> None:
    """Validate the given action definition's `saliences` field.

    Args:
        action_definition: An action definition from a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The action definition has an invalid `saliences` field.
    """
    # Detect role entries for undefined roles
    all_role_names = get_all_action_role_names(action_definition=action_definition)
    for role_name in action_definition['saliences']['roles']:
        if role_name not in all_role_names:
            raise errors.VivCompileError(
                f"Action '{action_definition['name']}' has 'saliences' with 'roles' "
                f"entry for undefined role: '{role_name}'"
            )
    # Detect cases of a saliences variable shadowing the name of a role from the same action
    if action_definition['saliences']['variable']:
        if action_definition['saliences']['variable']['name'] in all_role_names:
            raise errors.VivCompileError(
                f"Action '{action_definition['name']}' has 'saliences' field with custom 'for _' variable name that "
                f"shadows role name '{action_definition['saliences']['variable']['name']}' (this is not allowed)"
            )
    # Detect cases of a custom salience-yielding expression that won't evaluate to a numeric value
    for custom_expression in action_definition['saliences']['custom']:
        if is_non_numeric_expression(expression=custom_expression):
            raise errors.VivCompileError(
                msg=(
                    f"Action '{action_definition['name']}' has 'saliences' field with custom 'for _' expression "
                    f"with invalid expression type (value will be non-numeric): "
                    f"{custom_expression['source']['code']}"
                ),
                source=custom_expression['source'],
            )


def _validate_action_associations(action_definition: external_types.ActionDefinition) -> None:
    """Validate the given action definition's `associations` field.

    Args:
        action_definition: An action definition from a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The action definition has an invalid `associations` field.
    """
    # Detect role entries for undefined roles
    all_role_names = get_all_action_role_names(action_definition=action_definition)
    for role_name in action_definition['associations']['roles']:
        if role_name not in all_role_names:
            raise errors.VivCompileError(
                f"Action '{action_definition['name']}' has 'associations' field with 'roles' "
                f"entry for undefined role: '{role_name}'"
            )
    # Detect cases of an associations variable shadowing the name of a role from the same action
    if action_definition['associations']['variable']:
        if action_definition['associations']['variable']['name'] in all_role_names:
            raise errors.VivCompileError(
                f"Action '{action_definition['name']}' has 'associations' field with 'custom' variable name that "
                f"shadows role name '{action_definition['associations']['variable']['name']}' (this is not allowed)"
            )



"""The visitor component of the Viv DSL compiler, composed by a number of thematic mixin classes."""

from arpeggio import PTNodeVisitor

from ._actions import VisitorMixinActions
from ._bindings import VisitorMixinBindings
from ._expressions import VisitorMixinExpressions
from ._files import VisitorMixinFiles
from ._plans import VisitorMixinPlans
from ._queries import VisitorMixinQueries
from ._reactions import VisitorMixinReactions
from ._roles import VisitorMixinRoles
from ._selectors import VisitorMixinSelectors
from ._sifting_patterns import VisitorMixinSiftingPatterns
from ._statements import VisitorMixinStatements
from ._temporal_constraints import VisitorMixinTemporalConstraints
from ._tropes import VisitorMixinTropes


class Visitor(
    VisitorMixinActions,
    VisitorMixinBindings,
    VisitorMixinExpressions,
    VisitorMixinFiles,
    VisitorMixinPlans,
    VisitorMixinQueries,
    VisitorMixinReactions,
    VisitorMixinRoles,
    VisitorMixinSelectors,
    VisitorMixinSiftingPatterns,
    VisitorMixinStatements,
    VisitorMixinTemporalConstraints,
    VisitorMixinTropes,
    PTNodeVisitor,  # Must come last, otherwise order doesn't matter (no duplicate method names)
):
    """The composed visitor class for Viv ASTs, containing methods for visiting all
    nonterminal symbols that appear in the Viv DSL grammar.
    """
    pass



"""Handles the construction of role-dependency forests.

These forests will be used at runtime to structure role casting during construct targeting,
where it drives an optimized procedure that helps to prune the space of possible casts
and thereby significantly reduce the degree to which conditions are tested.
"""

__all__ = ["build_role_dependency_forests", "build_condition_groups"]

from viv_compiler import external_types, internal_types, utils, validation


def build_role_dependency_forests(combined_ast: internal_types.CombinedAST) -> None:
    """Construct and attribute a role-dependency forest for each construct in the given combined AST.

    The trees constituting the forests constructed by this function will be defined by setting the `parent`
    and `children` fields in the applicable role definitions. Dependency forests are used at runtime to
    structure role casting during construct targeting. Edges represent dependency from two sources:

        1) The child role's pool directive references the parent role, which means the latter
           must be cast first. This is a strict dependency.

        2) The child role shares one or more conditions with the parent role and/or its ancestors
           in the tree (as constructed to that point). This is not a strict dependency, but rather
           one that enables significant optimizations.

    A role-dependency forest is composed of one or more role-dependency trees. Within a given tree are
    one or more roles structured such that each child depends on its parent role. If there are multiple
    trees for a given construct, the roles in the respective trees are strictly independent of one another,
    in terms of the two forms of dependency enumerated above.

    During construct targeting, role casting iterates over each dependency tree in the construct's forest,
    traversing each in a depth-first manner, with backtracking working in the inverse direction. This allows
    for sequential role casting and also greatly reduces the frequency of re-evaluating conditions unnecessarily,
    since backtracking will not revisit roles with no dependency relation to the one for which casting failed.

    Finally, note that this tree does not contain optional roles (ones with a `min` of `0`) because they
    are always cast last, after all required roles have been cast, in a manner that does not entail
    backtracking. Further, by the time we get here, we have already confirmed that no casting-pool
    declarations reference an optional role, so no role can strictly depend on an optional role.

    Args:
        combined_ast: A combined AST with wrapped expressions and actions that have the `initiator` field set.

    Returns:
        None (modifies the construct definitions in place).
    """
    # Conduct preliminary validation of the roles in the current intermediate construct
    # definitions, to catch any issues that would preclude construction of the forests.
    validation.prevalidate_roles(combined_ast=combined_ast)
    # If that passes, we can construct a forest for each construct definition in turn
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        # Isolate the required roles, since only these can serve as nodes in the forest
        required_roles = [role for role in construct_definition['roles'].values() if role['min'] > 0]
        # Handle dependency relations that are rooted in casting-pool directives
        _handle_casting_pool_dependencies(construct_definition=construct_definition, required_roles=required_roles)
        # Handle dependency relations rooted in shared conditions
        _handle_condition_dependencies(construct_definition=construct_definition, required_roles=required_roles)
        # Finally, organize the dependency forest we have constructed
        _organize_dependency_forest(construct_definition=construct_definition, required_roles=required_roles)


def _handle_casting_pool_dependencies(
    construct_definition: internal_types.IntermediateConstructDefinition,
    required_roles: list[external_types.RoleDefinition]
) -> None:
    """Modify the given construct definition in place to attribute dependency relations among
    its roles that are rooted in their respective casting-pool directives.

    If a given role R1 references a single other role R2 in its casting-pool directive, R1 will be specified
    as a child of the 'anchor role' R2, in the dependency tree. If there are multiple such anchor roles for R1,
    only one of them will become R1's parent, but all of them will be ancestors.

    Note that by this point, we will have already caught cases of the following:
        - A cycle among the casting-pool directives.
        - A casting-pool directive anchored in an optional role.
        - An initiator role with a casting pool.

    Args:
        construct_definition: An intermediate construct definition.
        required_roles: A list containing definitions for all required roles in the given construct definition.

    Returns:
        None (modifies the construct definition in place).
    """
    for role_definition in required_roles:
        if not role_definition['pool']:
            continue
        anchor_role_names = utils.get_all_referenced_roles(ast_chunk=role_definition['pool'])
        if construct_definition['type'] == external_types.ConstructDiscriminator.ACTION:
            # If this is an action, we want to ensure that the initiator always dominates, which
            # can be accomplished by placing them last here. This works because we iterate over
            # the anchor role names in order, reattaching to each in turn.
            anchor_role_names.sort(key=lambda role_name: role_name == construct_definition['initiator'])
        for anchor_role_name in anchor_role_names:
            _reattach_subtree(
                construct_definition=construct_definition,
                reattachment_target_role_name=anchor_role_name,
                reattachment_source_role_name=role_definition['name']
            )


def _handle_condition_dependencies(
    construct_definition: internal_types.IntermediateConstructDefinition,
    required_roles: list[external_types.RoleDefinition]
) -> None:
    """Modify the given construct definition in place to attribute dependency relations among
    its roles that are rooted in shared conditions.

    Args:
        construct_definition: An intermediate construct definition with wrapped expressions. If it's an
             action, it must also have the `initiator` field set.
        required_roles: A list containing definitions for all required roles in the given construct definition.

    Returns:
        None (modifies the construct definition in place).
    """
    # Compile all cases of roles sharing conditions
    role_name_to_role_names_sharing_conditions = {role['name']: set() for role in required_roles}
    for condition in construct_definition["_conditions_raw"]:
        if any(construct_definition['roles'][role_name]['min'] == 0 for role_name in condition['references']):
            # Skip any condition that references an optional role
            continue
        for condition_role_name in condition["references"]:
            for other_condition_role_name in condition["references"]:
                if condition_role_name == other_condition_role_name:
                    continue
                role_name_to_role_names_sharing_conditions[condition_role_name].add(other_condition_role_name)
    # For each role R, retrieve each other role S with which R shares one or more conditions. If S is not
    # a lineal relative of R, we will reattach under either R or S a subtree containing the other.
    role_definition_order = {role_name: i for i, role_name in enumerate(construct_definition['roles'])}
    for role_definition in required_roles:
        role_name = role_definition['name']
        for other_role_name in role_name_to_role_names_sharing_conditions[role_name]:
            if role_definition_order[other_role_name] < role_definition_order[role_name]:
                # The function `_linearize_condition_dependency()`, which we call below, requires that the
                # two roles be passed to certain parameters depending on which one was defined first by the
                # author. This is because we use definition order as a heuristic. The `role_definition_order`
                # mapping works because the keys in the `roles` field always appear in the author-defined
                # order (note that dictionary keys have been ordered since Python 3.7).
                continue
            role_ancestors = _get_ancestors(construct_definition=construct_definition, role_name=role_name)
            role_descendants = _get_descendants(construct_definition=construct_definition, role_name=role_name)
            if other_role_name in role_ancestors or other_role_name in role_descendants:
                continue
            # Linearize the dependency
            reattachment_target_role_name, reattachment_source_role_name = _linearize_condition_dependency(
                construct_definition=construct_definition,
                role_name=role_name,
                other_role_name=other_role_name
            )
            # Update the tree to reflect the linearized dependency
            _reattach_subtree(
                construct_definition=construct_definition,
                reattachment_target_role_name=reattachment_target_role_name,
                reattachment_source_role_name=reattachment_source_role_name
            )


def _linearize_condition_dependency(
    construct_definition: internal_types.IntermediateConstructDefinition,
    role_name: external_types.RoleName,
    other_role_name: external_types.RoleName
) -> tuple[external_types.RoleName, external_types.RoleName]:
    """Linearize the mutual condition dependency between the given roles.

    The given roles share a condition, which is a mutual dependency that we need to linearize,
    because we are building a tree. Since this is a mutual dependency, we could reasonably go in
    either direction, but for optimization purposes we will use three heuristics:

    1. The initiator dominates. We want them to be the root of their tree always.

    2. Otherwise, the role with the fewest ancestors wins.

    3. If there is a tie, the role defined first will win. Authors tend to write roles in decreasing
       order of importance, and more important roles tend to depend less on other roles, which makes
       for a good heuristic here.

    Args:
        role_name: The name of the role defined earlier in the construct definition.
        other_role_name: The name of the role defined later in the construct definition.

    Returns:
        A tuple containing, in order, the ancestor and the descendant in the linearized dependency.
    """
    # If this is an action, isolate the initiator role name, which is needed below
    initiator_name = None
    if construct_definition['type'] == external_types.ConstructDiscriminator.ACTION:
        initiator_name = construct_definition['initiator']
    # To break a potential tie, per heuristic #3 above, we'll declare as the winner
    # the role that was defined first, until proven otherwise.
    linearized_dependency_ancestor = role_name
    if initiator_name and role_name == initiator_name:
        linearized_dependency_ancestor = role_name
    elif initiator_name and other_role_name == initiator_name:
        linearized_dependency_ancestor = other_role_name
    else:
        role_ancestors = _get_ancestors(construct_definition=construct_definition, role_name=role_name)
        other_role_ancestors = _get_ancestors(construct_definition=construct_definition, role_name=other_role_name)
        if len(other_role_ancestors) < len(role_ancestors):  # Strictly fewer, so we can break ties per heuristic #3
            linearized_dependency_ancestor = other_role_name
    linearized_dependency_descendant = other_role_name if linearized_dependency_ancestor == role_name else role_name
    return linearized_dependency_ancestor, linearized_dependency_descendant


def _reattach_subtree(
    construct_definition: internal_types.IntermediateConstructDefinition,
    reattachment_target_role_name: str,
    reattachment_source_role_name: str
) -> None:
    """Modify the given construct definition in place to reattach a role-dependency subtree containing
    the given source role, such that the given target role will become one of its ancestors.

    If the attachment target is already an ancestor of the attachment source, no modifications
    are needed. Otherwise, we will climb up from the source to its highest ancestor A that is not
    also an ancestor of the target, and then reattach to the target the subtree rooted in A. We do
    this simply by changing A's parent to the target (and updating children records as needed).

    Naively, we might try to make the source's primogenitor, P, a child of the target, but this will
    fail in the case where the target also descends from P, in particular the case where the target's
    pool is anchored in P. (Since P would then be a child of the target, despite the target requiring
    P to be cast first.) In other words, we need to take special care when the source is a collateral
    relative of the target, hence our procedure here.

    Args:
        construct_definition: An intermediate construct definition for which forest construction is already underway.
        reattachment_target_role_name: The role that will become the parent of the reattached subtree.
        reattachment_source_role_name: The role whose ancestor will serve as the root of the reattached subtree.

    Returns:
        None (modifies the given construct definition in place).
    """
    # Climb up from the source to find to its highest ancestor `A` that is not also an ancestor
    # of the target. `A` will serve as the root of the subtree we reattach under the target.
    head_ancestor_names = _get_ancestors(
        construct_definition=construct_definition,
        role_name=reattachment_target_role_name
    )
    source_ancestor_definition: external_types.RoleDefinition = (
        construct_definition['roles'][reattachment_source_role_name]
    )
    while source_ancestor_definition['parent']:
        if source_ancestor_definition['parent'] == reattachment_target_role_name:
            # No modifications are needed, because the source is already a descendant of the target
            return
        if source_ancestor_definition['parent'] in head_ancestor_names:
            # We found `A`, so we can stop climbing
            break
        source_ancestor_definition = construct_definition['roles'][source_ancestor_definition['parent']]
    # Leaving its descendant structure intact, make `A` a child of the target
    if source_ancestor_definition['parent']:
        original_parent = construct_definition['roles'][source_ancestor_definition['parent']]
        original_parent['children'].remove(source_ancestor_definition['name'])
    source_ancestor_definition['parent'] = reattachment_target_role_name
    reattachment_head_role_definition = construct_definition['roles'][reattachment_target_role_name]
    reattachment_head_role_definition['children'].append(source_ancestor_definition['name'])


def _organize_dependency_forest(
    construct_definition: internal_types.IntermediateConstructDefinition,
    required_roles: list[external_types.RoleDefinition]
) -> None:
    """Organize the dependency forest for the given construct definition, and attribute its forest roots.

    Args:
        construct_definition: An intermediate construct definition for which forest construction has already occurred.
        required_roles: A list containing definitions for all required roles in the given construct definition.

    Returns:
        None (modifies the construct definition in place).
    """
    # While we've constructed the forest by attributing parents and children to all the required roles,
    # we haven't stored any listing of the trees formed thereby, or specifically their roots. We need
    # such a listing so that we can proceed through the trees one by one during role casting, to cast
    # the roles in a depth-first manner.
    role_forest_tree_roots = [role['name'] for role in required_roles if not role['parent']]
    # Now we'll sort the forest tree roots such that larger trees come first. This heuristic proceeds
    # from an assumption that it's better to reach fail states as soon as possible, and fail states
    # are more likely to occur amid complex dependencies.
    role_forest_tree_roots.sort(
        key=lambda tree_root_name: len(_get_descendants(
            construct_definition=construct_definition,
            role_name=tree_root_name
        )),
        reverse=True
    )
    # Within each tree, let's also sort subtrees using the same heuristic
    for role_definition in construct_definition['roles'].values():
        role_definition['children'].sort(
            key=lambda child_role_name: len(_get_descendants(
                construct_definition=construct_definition,
                role_name=child_role_name
            )),
            reverse=True
        )
    # Finally, attribute the forest roots to the construct definition
    construct_definition['roleForestRoots'] = role_forest_tree_roots


def _get_ancestors(
    construct_definition: internal_types.IntermediateConstructDefinition,
    role_name: external_types.RoleName
) -> set[external_types.RoleName]:
    """Return the names of all ancestors of the given role in the dependency forest
    for the given construct definition.

    Args:
        construct_definition: An intermediate construct definition for which forest construction is already underway.
        role_name: Name of the role whose dependency ancestors are to be retrieved.

    Returns:
        A set containing the names of all dependency ancestors of the given role.
    """
    role_ancestors: set[external_types.RoleName] = set()
    role_definition: external_types.RoleDefinition = construct_definition['roles'][role_name]
    parent_name: external_types.RoleName | None = role_definition['parent']
    if parent_name is not None:
        role_ancestors.add(parent_name)
        role_ancestors |= _get_ancestors(construct_definition=construct_definition, role_name=parent_name)
    return role_ancestors


def _get_descendants(
    construct_definition: internal_types.IntermediateConstructDefinition,
    role_name: external_types.RoleName
) -> set[external_types.RoleName]:
    """Return the names of all descendants of the given role in the dependency forest
    for the given construct definition.

    Args:
        construct_definition: An intermediate construct definition for which forest construction is already underway.
        role_name: The role whose dependency descendants are to be retrieved.

    Returns:
        A set containing the names of all dependency descendants of the given role.
    """
    role_descendants: set[external_types.RoleName] = set()
    role_definition: external_types.RoleDefinition = construct_definition['roles'][role_name]
    for child_name in role_definition['children']:
        role_descendants.add(child_name)
        role_descendants |= _get_descendants(construct_definition=construct_definition, role_name=child_name)
    return role_descendants


def build_condition_groups(combined_ast: internal_types.CombinedAST) -> None:
    """Modify the given construct definitions in place by attributing a `conditions` field
    that maps role names to (only) the conditions that must be evaluated to cast that role.

    This function also sorts condition references in dependency-stream order.

    Note that by this point, we will have already caught cases of a condition referencing multiple optional roles.

    Args:
        combined_ast: A combined AST following role-forest construction.

    Returns:
        None (modifies the construct definitions in place).
    """
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        global_conditions = []  # See note just below
        role_name_to_conditions = {role_name: [] for role_name in construct_definition['roles']}
        for condition in construct_definition["_conditions_raw"]:
            # If there are no references, assign it to the global conditions. Such conditions
            # (e.g., a bare chance expression) can be evaluated at the very beginning of targeting,
            # prior to casting any roles.
            if not condition['references']:
                global_conditions.append(condition)
                continue
            # If there's a single reference, assign it to that role
            if len(condition['references']) == 1:
                role_name_to_conditions[condition['references'][0]].append(condition)
                continue
            # If there's an optional role referenced, assign it to the optional role. Note that by
            # now we have already enforced that a condition reference at most one optional role.
            assigned_to_optional = False
            for role_name in condition['references']:
                if construct_definition['roles'][role_name]['min'] == 0:
                    role_name_to_conditions[role_name].append(condition)
                    assigned_to_optional = True
                    break
            if assigned_to_optional:
                continue
            # Otherwise, assign it to the role furthest downstream in the dependency structure. Since all
            # roles sharing a condition will be situated in a direct line in this structure, we can
            # easily identify the most downstream role as the one with the most ancestors. For clarity,
            # we'll sort the actual condition `references` value in upstream-to-downstream order.
            condition['references'].sort(
                key=lambda role_name: len(_get_ancestors(
                    construct_definition=construct_definition,
                    role_name=role_name
                ))
            )
            role_name_to_conditions[condition['references'][-1]].append(condition)
        construct_conditions = external_types.ConstructConditions(
            globalConditions=global_conditions,
            roleConditions=role_name_to_conditions
        )
        construct_definition["conditions"] = construct_conditions



"""Handles the creation of metadata to include in a compiled content bundle."""

__all__ = ["create_metadata"]

from typing import Any

from viv_compiler import __version__, __schema_version__, __grammar_version__, external_types, internal_types, utils


def create_metadata(combined_ast: internal_types.CombinedAST) -> external_types.ContentBundleMetadata:
    """Return a package containing metadata for a compiled content bundle, derived from the given combined AST.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A metadata package for a compiled content bundle, derived from the given combined AST.
    """
    # Compile all referenced enums
    referenced_enum_names = _get_all_referenced_enum_names(ast_chunk=combined_ast)
    # Compile all referenced custom functions
    referenced_custom_function_names = _get_all_referenced_custom_function_names(ast_chunk=combined_ast)
    # Compile all reactions constrained by time of day
    time_of_day_parameterized_reactions = _get_time_of_day_parameterized_reactions(combined_ast=combined_ast)
    # Compile all queries referencing time of day
    time_of_day_parameterized_queries = _get_time_of_day_parameterized_queries(combined_ast=combined_ast)
    # Determine whether there is at least one assignment that modifies entity data
    found_entity_data_assignment = any(
        _has_entity_data_assignments(construct_definition=construct_definition)
        for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast)
    )
    # Package up the metadata and return it
    metadata = external_types.ContentBundleMetadata(
        schemaVersion=__schema_version__,
        compilerVersion=__version__,
        grammarVersion=__grammar_version__,
        referencedEnums=referenced_enum_names,
        referencedFunctionNames=referenced_custom_function_names,
        timeOfDayParameterizedReactions=time_of_day_parameterized_reactions,
        timeOfDayParameterizedQueries=time_of_day_parameterized_queries,
        hasEntityDataAssignments=found_entity_data_assignment
    )
    return metadata


def _get_all_referenced_enum_names(ast_chunk: Any) -> list[external_types.EnumName]:
    """Return the names of all enums referenced in the given AST chunk.

    This list will be stored in the compiled content bundle, where it's used for
    validation purposes upon the initialization of a Viv runtime.

    Args:
        ast_chunk: The full or partial AST to search for enum references.

    Returns:
        A list containing the names of all the enums referenced in the given AST chunk.
    """
    all_enum_expressions: list[external_types.EnumValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ENUM,
        ast_chunk=ast_chunk
    )
    all_referenced_enum_names = {expression['name'] for expression in all_enum_expressions}
    return sorted(all_referenced_enum_names)


def _get_all_referenced_custom_function_names(ast_chunk: Any) -> list[external_types.CustomFunctionName]:
    """Return the names of all custom functions referenced in the given AST chunk.

    This list will be stored in the compiled content bundle, where it's used for
    validation purposes upon the initialization of a Viv runtime.

    Args:
        ast_chunk: The full or partial AST to search for custom-function references.

    Returns:
        A list containing the names of all the custom functions referenced in the given AST chunk.
    """
    all_custom_function_references: list[external_types.CustomFunctionCallValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.CUSTOM_FUNCTION_CALL,
        ast_chunk=ast_chunk
    )
    all_referenced_custom_function_names = {expression['name'] for expression in all_custom_function_references}
    return sorted(all_referenced_custom_function_names)


def _get_time_of_day_parameterized_reactions(
    combined_ast: internal_types.CombinedAST
) -> list[external_types.TimeOfDayParameterizedReaction]:
    """Return a list specifying all reactions that are parameterized by time of day.

    This list will be stored in the compiled content bundle, where it's used for
    validation purposes upon the initialization of a Viv runtime.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A list specifying all reactions that are parameterized by time of day.
    """
    time_of_day_parameterized_reactions = []
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        all_nested_reactions: list[external_types.ReactionValue] = utils.get_all_expressions_of_type(
            expression_type=external_types.ExpressionDiscriminator.REACTION,
            ast_chunk=construct_definition
        )
        for reaction in all_nested_reactions:
            if reaction["time"]:
                for temporal_constraint in reaction["time"]:
                    if temporal_constraint["type"] == external_types.TemporalStatementDiscriminator.TIME_OF_DAY:
                        time_of_day_parameterized_reaction = external_types.TimeOfDayParameterizedReaction(
                            constructType=construct_definition["type"],
                            constructName=construct_definition["name"],
                            reaction=reaction["targetName"]
                        )
                        time_of_day_parameterized_reactions.append(time_of_day_parameterized_reaction)
    return time_of_day_parameterized_reactions


def _get_time_of_day_parameterized_queries(
    combined_ast: internal_types.CombinedAST
) -> list[external_types.QueryName]:
    """Return a list specifying all queries that are parameterized by time of day.

    This list will be stored in the compiled content bundle, where it's used for
    validation purposes upon the initialization of a Viv runtime.

    Args:
        combined_ast: A combined AST that has undergone all postprocessing.

    Returns:
        A list specifying all queries that are parameterized by time of day.
    """
    time_of_day_parameterized_queries: list[external_types.QueryName] = []
    for query_definition in combined_ast["queries"]:
        if query_definition['time']:
            for temporal_constraint in query_definition["time"]:
                if temporal_constraint["type"] == external_types.TemporalStatementDiscriminator.TIME_OF_DAY:
                    time_of_day_parameterized_queries.append(query_definition['name'])
    return time_of_day_parameterized_queries


def _has_entity_data_assignments(construct_definition: internal_types.IntermediateConstructDefinition) -> bool:
    """Return whether the given construct definition has at least one assignment that modifies entity data.

    Args:
        construct_definition: The construct definition in question.

    Returns:
        Whether the given construct definition has at least one assignment that modifies entity data.
    """
    all_assignments: list[external_types.AssignmentValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ASSIGNMENT,
        ast_chunk=construct_definition
    )
    for assignment in all_assignments:
        # Any entity-reference assignment modifies entity data
        if assignment["left"]["type"] == external_types.ExpressionDiscriminator.ENTITY_REFERENCE:
            return True
        # A symbol-reference assignment modifies entity data if it includes a pointer,
        # since pointers always target entity data, by definition.
        assignment_lhs: external_types.ReferenceValue = assignment["left"]["value"]
        for component in assignment_lhs["path"]:
            if (
                component["type"] ==
                external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_POINTER
            ):
                return True
    return False



"""Mixin class containing visitor methods associated with selector definitions."""

__all__ = ["VisitorMixinSelectors"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, sentinels, utils, validation


class VisitorMixinSelectors(PTNodeVisitor):
    """A visitor mixin for Viv action selectors and plan selectors."""

    @staticmethod
    def visit_selector(
        node: NonTerminal,
        children: list[Any]
    ) -> internal_types.IntermediateActionSelectorDefinition | internal_types.IntermediatePlanSelectorDefinition:
        """Visit a <selector> node."""
        # Populate the fields that were authored
        accumulated_fields = {}
        selector_header, selector_body = children
        accumulated_fields.update(selector_header)
        all_role_definitions = []
        if "roles" in selector_body:
            all_role_definitions += selector_body["roles"]
        validation.prevalidate_role_names(
            construct_type=accumulated_fields['type'],
            construct_name=accumulated_fields['name'],
            role_definitions=all_role_definitions,
            source=utils.derive_source_annotations(node=node)
        )
        accumulated_fields["roles"] = {role["name"]: role for role in all_role_definitions}
        for field_name, field_value in selector_body.items():
            if field_name == "roles":
                continue  # Already handled just above
            else:
                accumulated_fields[field_name] = field_value
        if "candidates" not in accumulated_fields:
            error_message = f"Selector '{accumulated_fields['name']}' is missing 'target' field (required)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        # Package up the intermediate definition
        if accumulated_fields["type"] == external_types.ConstructDiscriminator.ACTION_SELECTOR:
            intermediate_selector_definition = internal_types.IntermediateActionSelectorDefinition(
                type=external_types.ConstructDiscriminator.ACTION_SELECTOR,
                name=accumulated_fields["name"],
                roles=accumulated_fields["roles"],
                _conditions_raw=accumulated_fields.get("_conditions_raw", []),
                reserved=accumulated_fields.get("reserved", False),
                policy=accumulated_fields["policy"],
                candidates=accumulated_fields["candidates"]
            )
        else:
            if "reserved" in accumulated_fields:
                raise errors.VivCompileError(
                    f"Plan selector '{accumulated_fields['name']}' is marked 'reserved', but this is prohibited "
                    f"(plans and plan selectors are always reserved, so the marker is not needed)",
                    source=utils.derive_source_annotations(node=node)
                )
            intermediate_selector_definition = internal_types.IntermediatePlanSelectorDefinition(
                type=external_types.ConstructDiscriminator.PLAN_SELECTOR,
                name=accumulated_fields["name"],
                roles=accumulated_fields["roles"],
                _conditions_raw=accumulated_fields.get("_conditions_raw", []),
                policy=accumulated_fields["policy"],
                candidates=accumulated_fields["candidates"]
            )
        return intermediate_selector_definition

    @staticmethod
    def visit_selector_header(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <selector_header> node."""
        accumulated_fields = {}
        for child in children:
            if isinstance(child, str):
                accumulated_fields['name'] = child
            else:
                accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_selector_type(
        _, children: list[dict[str, external_types.ConstructDiscriminator]]
    ) -> dict[str, external_types.ConstructDiscriminator]:
        """Visit a <selector_type> node."""
        return children[0]

    @staticmethod
    def visit_action_selector_type(_, __) -> dict[str, external_types.ConstructDiscriminator]:
        """Visit an <action_selector_type> node."""
        return {"type": external_types.ConstructDiscriminator.ACTION_SELECTOR}

    @staticmethod
    def visit_plan_selector_type(_, __) -> dict[str, external_types.ConstructDiscriminator]:
        """Visit a <plan_selector_type> node."""
        return {"type": external_types.ConstructDiscriminator.PLAN_SELECTOR}

    @staticmethod
    def visit_selector_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <selector_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_selector_roles(
        _, children: list[external_types.RoleDefinition]
    ) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <selector_roles> node."""
        return {"roles": children}

    @staticmethod
    def visit_selector_conditions(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <selector_conditions> node."""
        return {"_conditions_raw": children[0]}

    @staticmethod
    def visit_selector_target_group(
        node: NonTerminal,
        children: list[internal_types.Sentinel | list[external_types.SelectorCandidate]]
    ) -> dict[str, Any]:
        """Visit an <selector_target_group> node."""
        selector_policy_sentinel, selector_candidates = children
        match selector_policy_sentinel:
            case sentinels.SELECTOR_POLICY_RANDOM:
                selector_policy = external_types.SelectorPolicy.RANDOMIZED
            case sentinels.SELECTOR_POLICY_WEIGHTED:
                selector_policy = external_types.SelectorPolicy.WEIGHTED
            case sentinels.SELECTOR_POLICY_ORDERED:
                selector_policy = external_types.SelectorPolicy.ORDERED
            case _:
                raise errors.VivCompileError(
                    f"Unexpected selector policy",
                    source=utils.derive_source_annotations(node=node)
                )
        accumulated_fields = {
            "policy": selector_policy,
            "candidates": selector_candidates
        }
        return accumulated_fields

    @staticmethod
    def visit_selector_policy(node: NonTerminal, children: list[str]) -> internal_types.Sentinel:
        """Visit an <selector_policy> node."""
        match children[0]:
            case "randomly":
                return sentinels.SELECTOR_POLICY_RANDOM
            case "with weights":
                return sentinels.SELECTOR_POLICY_WEIGHTED
            case "in order":
                return sentinels.SELECTOR_POLICY_ORDERED
            case _:
                raise errors.VivCompileError(
                    f"Unexpected selector policy",
                    source=utils.derive_source_annotations(node=node)
                )

    @staticmethod
    def visit_selector_candidates(
        _, children: list[external_types.SelectorCandidate]
    ) -> list[external_types.SelectorCandidate]:
        """Visit an <selector_candidates> node."""
        return children

    @staticmethod
    def visit_selector_candidate(_, children: list[Any]) -> external_types.SelectorCandidate:
        """Visit an <selector_candidate> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        selector_candidate = external_types.SelectorCandidate(
            name=accumulated_fields['name'],
            isSelector=accumulated_fields['isSelector'],
            bindings=accumulated_fields.get('bindings', external_types.PrecastBindings(partial=True, roles={})),
            weight=accumulated_fields.get('weight', None)
        )
        return selector_candidate

    @staticmethod
    def visit_selector_candidate_name(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <selector_candidate_name> node."""
        accumulated_fields = {
            "name": children[-1],
            "isSelector": len(children) > 1
        }
        return accumulated_fields

    @staticmethod
    def visit_selector_candidate_weight(
        _, children: list[external_types.Expression]
    ) -> dict[str, external_types.Expression]:
        """Visit an <selector_candidate_weight> node."""
        return {"weight": children[0]}



"""Mixin class containing visitor methods associated with query definitions."""

from __future__ import annotations

__all__ = ["VisitorMixinQueries"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, utils, validation


class VisitorMixinQueries(PTNodeVisitor):
    """A visitor mixin for Viv queries."""

    @staticmethod
    def visit_query(node: NonTerminal, children: list[Any]) -> internal_types.IntermediateQueryDefinition:
        """Visit a <query> node."""
        # Populate the fields that were authored
        accumulated_fields = {}
        query_name, query_body = children
        for field_name, field_value in query_body.items():
            if field_name == "roles":
                validation.prevalidate_role_names(
                    construct_type=external_types.ConstructDiscriminator.QUERY,
                    construct_name=query_name,
                    role_definitions=field_value,
                    source=utils.derive_source_annotations(node=node)
                )
                accumulated_fields["roles"] = {role["name"]: role for role in field_value}
            else:
                accumulated_fields[field_name] = field_value
        # Package up the intermediate definition
        intermediate_query_definition = internal_types.IntermediateQueryDefinition(
            type=external_types.ConstructDiscriminator.QUERY,
            name=query_name,
            roles=accumulated_fields.get("roles", {}),
            _conditions_raw=accumulated_fields.get("_conditions_raw", []),
            actionName=accumulated_fields.get("actionName", None),
            ancestors=accumulated_fields.get("ancestors", None),
            descendants=accumulated_fields.get("descendants", None),
            importance=accumulated_fields.get("importance", None),
            tags=accumulated_fields.get("tags", None),
            salience=accumulated_fields.get("salience", None),
            associations=accumulated_fields.get("associations", None),
            location=accumulated_fields.get("location", None),
            time=accumulated_fields.get("time", None),
            initiator=accumulated_fields.get("initiator", None),
            partners=accumulated_fields.get("partners", None),
            recipients=accumulated_fields.get("recipients", None),
            bystanders=accumulated_fields.get("bystanders", None),
            active=accumulated_fields.get("active", None),
            present=accumulated_fields.get("present", None)
        )
        return intermediate_query_definition

    @staticmethod
    def visit_query_header(_, children: list[external_types.QueryName]) -> external_types.QueryName:
        """Visit a <query_header> node."""
        return children[0]

    @staticmethod
    def visit_query_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <query_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_query_roles(
        _, children: list[external_types.RoleDefinition]
    ) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <query_roles> node."""
        return {"roles": children}

    @staticmethod
    def visit_query_conditions(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <query_conditions> node."""
        return {"_conditions_raw": children[0]}

    @staticmethod
    def visit_query_action_name(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_action_name> node."""
        return {"actionName": children}

    @staticmethod
    def visit_query_ancestors(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_ancestors> node."""
        return {"ancestors": children}

    @staticmethod
    def visit_query_descendants(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_descendants> node."""
        return {"descendants": children}

    @staticmethod
    def visit_query_importance(
        _, children: list[external_types.QueryNumericRange]
    ) -> dict[str, external_types.QueryNumericRange]:
        """Visit a <query_importance> node."""
        return {"importance": children[0]}

    @staticmethod
    def visit_query_tags(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_tags> node."""
        return {"tags": children}

    @staticmethod
    def visit_query_salience(
        _, children: list[external_types.QueryNumericRange]
    ) -> dict[str, external_types.QueryNumericRange]:
        """Visit a <query_salience> node."""
        return {"salience": children[0]}

    @staticmethod
    def visit_query_associations(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_associations> node."""
        return {"associations": children}

    @staticmethod
    def visit_query_location(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_location> node."""
        return {"location": children}

    @staticmethod
    def visit_query_time(
        _, children: list[external_types.TemporalConstraint]
    ) -> dict[str, list[external_types.TemporalConstraint]]:
        """Visit a <query_time> node."""
        return {"time": children}

    @staticmethod
    def visit_query_initiator(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_initiator> node."""
        return {"initiator": children}

    @staticmethod
    def visit_query_partners(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_partners> node."""
        return {"partners": children}

    @staticmethod
    def visit_query_recipients(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_recipients> node."""
        return {"recipients": children}

    @staticmethod
    def visit_query_bystanders(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_bystanders> node."""
        return {"bystanders": children}

    @staticmethod
    def visit_query_active(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_active> node."""
        return {"active": children}

    @staticmethod
    def visit_query_present(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <query_present> node."""
        return {"present": children}

    @staticmethod
    def visit_query_numeric_criteria(node: NonTerminal, children: list[Any]) -> external_types.QueryNumericRange:
        """Visit a <query_numeric_criteria> node."""
        # Prevalidate operators before we lose this information
        operators = ("==", ">=", ">", "<", "<=")
        counts = {operator: 0 for operator in operators}
        for operator, _ in children:
            counts[operator] += 1
        if counts["=="] and len(children) > 1:
            error_message = f"When '==' is used in a numeric query field, no other criterion may be present"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        for operator in operators:
            if counts[operator] > 1:
                error_message = f"Operator '{operator}' is used multiple times in a numeric query field"
                raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if counts[">"] and counts[">="]:
            error_message = (
                f"Operators '>' and '>=' are both used in a numeric query field (only one may be used)"
            )
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if counts["<"] and counts["<="]:
            error_message = (
                f"Operators '<' and '<=' are both used in a numeric query field (only one may be used)"
            )
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        # Construct the implied range to populate the component
        upper, lower = None, None
        for operator, operand in children:
            if operator in (">", ">=", "=="):
                lower = external_types.QueryNumericRangeBound(value=operand, inclusive=operator != ">")
            if operator in ("<", "<=", "=="):
                upper = external_types.QueryNumericRangeBound(value=operand, inclusive=operator != "<")
        query_numeric_range = external_types.QueryNumericRange(lower=lower, upper=upper)
        return query_numeric_range

    @staticmethod
    def visit_query_numeric_criterion(
        _, children: list[str | external_types.SalienceScoreExpression]
    ) -> list[str | external_types.SalienceScoreExpression]:
        """Visit a <query_numeric_criterion> node."""
        return children

    @staticmethod
    def visit_query_numeric_criterion_operator(node: NonTerminal, _) -> str:
        """Visit a <query_numeric_criterion_operator> node."""
        return str(node)

    @staticmethod
    def visit_set_predicate(_, children: list[Any]) -> external_types.SetPredicate:
        """Visit a <set_predicate> node."""
        operator, *operand = children
        set_predicate = external_types.SetPredicate(operator=operator, operand=operand)
        return set_predicate

    @staticmethod
    def visit_set_predicate_operator(node: NonTerminal, _) -> str:
        """Visit a <set_predicate_operator> node."""
        return str(node)

    @staticmethod
    def visit_set_predicate_tags(_, children: list[Any]) -> external_types.SetPredicate:
        """Visit a <set_predicate_tags> node."""
        return VisitorMixinQueries.visit_set_predicate(_, children)



"""Internal functionality for validating sifting-pattern definitions."""

__all__ = ["validate_sifting_pattern_definitions"]

from viv_compiler import external_types
from ._common import validate_common_concerns


def validate_sifting_pattern_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the sifting-pattern definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.
    """
    for sifting_pattern_definition in content_bundle["siftingPatterns"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=sifting_pattern_definition, content_bundle=content_bundle)



"""Viv Compiler: the reference compiler for the DSL at the heart of the Viv system for emergent narrative."""

__all__ = [
    "__version__",
    "__grammar_version__",
    "__schema_version__",
    "compile_from_path",
    "ContentBundle",
    "VivCompileError",
    "VivParseError"
]

from ._versions import __version__, __grammar_version__, __schema_version__
from .api import compile_from_path  # Relies on `__version__` already being set
from .errors import VivCompileError, VivParseError
from .external_types import ContentBundle
